#ifndef TFR_H
#define TFR_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <limits>
#include <libxml++/parsers/domparser.h>
#include <boost/logic/tribool.hpp>
#include <boost/logic/tribool_io.hpp>

#include "sysdeps.h"
#include "fplan.h"
#include "dbobj.h"
#include "engine.h"

class TrafficFlowRestrictions {
protected:
	typedef std::set<Glib::ustring> tracerules_t;

public:
        typedef Engine::AirwayGraphResult::Graph Graph;
        typedef Engine::AirwayGraphResult::Vertex Vertex;
        typedef Engine::AirwayGraphResult::Edge Edge;

	static const Graph::vertex_descriptor invalid_vertex_descriptor = ~0UL;

	class RuleWpt;
	class AltRange;

	class FplWpt : public FPlanWaypoint {
	public:
		typedef Vertex::type_t type_t;

		FplWpt(void) : m_vertex_descriptor(invalid_vertex_descriptor), m_wptnr(0), m_type(Vertex::type_undefined), m_navaid_type(NavaidsDb::Navaid::navaid_invalid) {}
		void set_vertex_descriptor(Graph::vertex_descriptor u = invalid_vertex_descriptor) { m_vertex_descriptor = u; }
		Graph::vertex_descriptor get_vertex_descriptor(void) const { return m_vertex_descriptor; }
		void set_wptnr(unsigned int w) { m_wptnr = w; }
		unsigned int get_wptnr(void) const { return m_wptnr; }
		void set_type(type_t t = Vertex::type_undefined) { m_type = t; }
		type_t get_type(void) const { return m_type; }
		void set_navaid_type(NavaidsDb::Navaid::navaid_type_t nt = NavaidsDb::Navaid::navaid_invalid) { m_navaid_type = nt; }
		NavaidsDb::Navaid::navaid_type_t get_navaid_type(void) const { return m_navaid_type; }
		bool has_ndb(void) const { return get_type() == Vertex::type_navaid && NavaidsDb::Navaid::has_ndb(get_navaid_type()); }
		bool has_vor(void) const { return get_type() == Vertex::type_navaid && NavaidsDb::Navaid::has_vor(get_navaid_type()); }
		bool has_dme(void) const { return get_type() == Vertex::type_navaid && NavaidsDb::Navaid::has_dme(get_navaid_type()); }
		const Glib::ustring& get_ident(void) const;
		bool operator==(const RuleWpt& r) const;
		bool within(const AltRange& ar) const;

	protected:
		Graph::vertex_descriptor m_vertex_descriptor;
		unsigned int m_wptnr;
		type_t m_type;
		NavaidsDb::Navaid::navaid_type_t m_navaid_type;
	};

	class Fpl : public std::vector<FplWpt> {
	public:
		Fpl(void) {}
		Fpl(const FPlanRoute& rte) { *this = rte; }
		Fpl& operator=(const FPlanRoute& rte);
		Rect get_bbox(void) const;
	};

	class Message {
	public:
		typedef std::set<unsigned int> set_t;
		typedef enum {
			type_error,
			type_warning,
			type_info,
			type_tracecond,
			type_tracetfe
		} type_t;

		Message(const Glib::ustring& text = "", type_t t = type_error, const Glib::ustring& rule = "") : m_text(text), m_rule(rule), m_type(t) {}
		const set_t& get_vertexset(void) const { return m_vertexset; }
		const set_t& get_edgeset(void) const { return m_edgeset; }
		const Glib::ustring& get_text(void) const { return m_text; }
		const Glib::ustring& get_rule(void) const { return m_rule; }
		type_t get_type(void) const { return m_type; }
		static const std::string& get_type_string(type_t t);
		const std::string& get_type_string(void) const { return get_type_string(get_type()); }
		static char get_type_char(type_t t);
		char get_type_char(void) const { return get_type_char(get_type()); }

	protected:
		set_t m_vertexset;
		set_t m_edgeset;
		Glib::ustring m_text;
		Glib::ustring m_rule;
		type_t m_type;

		friend class TrafficFlowRestrictions;
		void add_vertex(unsigned int nr);
		void add_edge(unsigned int nr);
		void insert_vertex(unsigned int nr);
	};

	class RuleResult {
	public:
		typedef std::set<unsigned int> set_t;

		class Alternative {
		public:

			class Sequence {
			public:
				typedef enum {
					type_airway,
					type_dct,
					type_point,
					type_sid,
					type_star,
					type_airspace,
					type_invalid
				} type_t;

				class RulePoint {
				public:
					RulePoint(const Glib::ustring& id = "", const Point& pt = Point(), Vertex::type_t typ = Vertex::type_undefined,
						  bool vor = false, bool ndb = false);
					RulePoint(const FplWpt& wpt, bool vor = false, bool ndb = false);
					RulePoint(const RuleWpt& wpt);
					RulePoint(const Vertex& vv);
					const Glib::ustring& get_ident(void) const { return m_ident; }
					const Point& get_coord(void) const { return m_coord; }
					Vertex::type_t get_type(void) const { return m_type; }
					bool is_vor(void) const { return m_vor; }
					bool is_ndb(void) const { return m_ndb; }

				protected:
					Glib::ustring m_ident;
					Point m_coord;
					Vertex::type_t m_type;
					bool m_vor;
					bool m_ndb;
				};

				Sequence(type_t t = type_dct, const RulePoint& pt0 = RulePoint(), const RulePoint& pt1 = RulePoint(),
					 int32_t lwralt = 0, int32_t upralt = std::numeric_limits<int32_t>::max(),
					 const Glib::ustring& id = "", const Glib::ustring& apt = "", char bc = 0)
					: m_point0(pt0), m_point1(pt1), m_ident(id), m_airport(apt),
					  m_lwralt(lwralt), m_upralt(upralt), m_bdryclass(bc), m_type(t) {}
				const RulePoint& get_point0(void) const { return m_point0; }
				const RulePoint& get_point1(void) const { return m_point1; }
				const Glib::ustring& get_ident(void) const { return m_ident; }
				const Glib::ustring& get_airport(void) const { return m_airport; }
				int32_t get_lower_alt(void) const { return m_lwralt; }
				int32_t get_upper_alt(void) const { return m_upralt; }
				char get_bdryclass(void) const { return m_bdryclass; }
				type_t get_type(void) const { return m_type; }
				static const std::string& get_type_string(type_t t);
				const std::string& get_type_string(void) const { return get_type_string(get_type()); }

			protected:
				RulePoint m_point0;
				RulePoint m_point1;
				Glib::ustring m_ident;
				Glib::ustring m_airport;
				int32_t m_lwralt;
				int32_t m_upralt;
				char m_bdryclass;
				type_t m_type;
			};

			Alternative(const set_t& vs = set_t(), const set_t& es = set_t()) : m_vertexset(vs), m_edgeset(es) {}
			unsigned int size(void) const { return m_seq.size(); }
			bool empty(void) const { return m_seq.empty(); }
			const Sequence& operator[](unsigned int idx) const { return m_seq[idx]; }
			const set_t& get_vertexset(void) const { return m_vertexset; }
			const set_t& get_edgeset(void) const { return m_edgeset; }

		protected:
			typedef std::vector<Sequence> seq_t;
			seq_t m_seq;
			set_t m_vertexset;
			set_t m_edgeset;

			friend class TrafficFlowRestrictions;
			void add_sequence(const Sequence& s) { m_seq.push_back(s); }
		};

		typedef enum {
			codetype_mandatory,
			codetype_forbidden,
			codetype_closed
		} codetype_t;

		RuleResult(const Glib::ustring& rule = "", const Glib::ustring& desc = "", const Glib::ustring& oprgoal = "",
			   codetype_t ct = codetype_closed, const set_t& vs = set_t(), const set_t& es = set_t())
			: m_vertexset(vs), m_edgeset(es), m_rule(rule), m_desc(desc), m_oprgoal(oprgoal), m_codetype(ct) {}
		unsigned int size(void) const { return m_elements.size(); }
		bool empty(void) const { return m_elements.empty(); }
		const Alternative& operator[](unsigned int idx) const { return m_elements[idx]; }
		const set_t& get_vertexset(void) const { return m_vertexset; }
		const set_t& get_edgeset(void) const { return m_edgeset; }
		const Glib::ustring& get_rule(void) const { return m_rule; }
		const Glib::ustring& get_desc(void) const { return m_desc; }
		const Glib::ustring& get_oprgoal(void) const { return m_oprgoal; }
		codetype_t get_codetype(void) const { return m_codetype; }

	protected:
		typedef std::vector<Alternative> elements_t;
		elements_t m_elements;
		set_t m_vertexset;
		set_t m_edgeset;
		Glib::ustring m_rule;
		Glib::ustring m_desc;
		Glib::ustring m_oprgoal;
		codetype_t m_codetype;

		friend class TrafficFlowRestrictions;
		void add_element(const Alternative& e) { m_elements.push_back(e); }
	};

	class Result {
	public:
		Result(const std::string& acft = "", const std::string& equip = "", char typeofflight = 'G');
		operator bool(void) const { return m_result; }
		unsigned int size(void) const { return m_rules.size(); }
		bool empty(void) const { return m_rules.empty(); }
		const RuleResult& operator[](unsigned int idx) const { return m_rules[idx]; }
		unsigned int messages_size(void) const { return m_messages.size(); }
		bool messages_empty(void) const { return m_messages.empty(); }
		const Message& get_message(unsigned int idx) const { return m_messages[idx]; }
		const Fpl& get_fplan(void) const { return m_fplan; }
		const std::string& get_aircrafttype(void) const { return m_aircrafttype; }
		const std::string& get_aircrafttypeclass(void) const { return m_aircrafttypeclass; }
		const std::string& get_equipment(void) const { return m_equipment; }
		char get_typeofflight(void) const { return m_typeofflight; }

	protected:
		typedef std::vector<RuleResult> rules_t;
		rules_t m_rules;
		typedef std::vector<Message> messages_t;
		messages_t m_messages;
		Fpl m_fplan;
		std::string m_aircrafttype;
		std::string m_aircrafttypeclass;
		std::string m_equipment;
		char m_typeofflight;
		bool m_result;

		friend class TrafficFlowRestrictions;
	};

	class RuleWpt {
	public:
		RuleWpt(FplWpt::type_t t = Vertex::type_undefined, const Glib::ustring& ident = "", const Point& coord = Point(), bool vor = false, bool ndb = false)
			: m_ident(ident), m_coord(coord), m_type(t), m_vor(vor), m_ndb(ndb) {}
		FplWpt::type_t get_type(void) const { return m_type; }
		const Glib::ustring& get_ident(void) const { return m_ident; }
		const Point& get_coord(void) const { return m_coord; }
		bool is_vor(void) const { return m_vor; }
		bool is_ndb(void) const { return m_ndb; }

	protected:
		Glib::ustring m_ident;
		Point m_coord;
		FplWpt::type_t m_type;
		bool m_vor;
		bool m_ndb;
	};

	class AltRange {
	public:
		typedef enum {
			alt_qnh,
			alt_std,
			alt_height
		} alt_t;

		AltRange(int32_t lwr = 0, alt_t lwrmode = alt_std, int32_t upr = std::numeric_limits<int32_t>::max(), alt_t uprmode = alt_std)
			: m_lwralt(lwr), m_upralt(upr), m_lwrmode(lwrmode), m_uprmode(uprmode) {}
		alt_t get_lower_mode(void) const { return m_lwrmode; }
		int32_t get_lower_alt(void) const { return m_lwralt; }
		alt_t get_upper_mode(void) const { return m_uprmode; }
		int32_t get_upper_alt(void) const { return m_upralt; }

	protected:
		int32_t m_lwralt;
		int32_t m_upralt;
		alt_t m_lwrmode;
		alt_t m_uprmode;
	};

	class RuleDct {
	public:
		RuleDct(Graph::vertex_descriptor u = invalid_vertex_descriptor,
			Graph::vertex_descriptor v = invalid_vertex_descriptor,
			const RuleWpt& w0 = RuleWpt(), const RuleWpt& w1 = RuleWpt()) : m_fromwpt(w0), m_towpt(w1), m_from(u), m_to(v) {}
		const RuleWpt& get_fromwpt(void) const { return m_fromwpt; }
		const RuleWpt& get_towpt(void) const { return m_towpt; }
		Graph::vertex_descriptor get_from(void) const { return m_from; }
		Graph::vertex_descriptor get_to(void) const { return m_to; }

	protected:
		RuleWpt m_fromwpt;
		RuleWpt m_towpt;
		Graph::vertex_descriptor m_from;
		Graph::vertex_descriptor m_to;
	};

	class RuleRoute : public RuleDct {
	public:
		RuleRoute(Graph::vertex_descriptor u = invalid_vertex_descriptor,
			  Graph::vertex_descriptor v = invalid_vertex_descriptor,
			  const RuleWpt& w0 = RuleWpt(), const RuleWpt& w1 = RuleWpt(),
			  const Glib::ustring& awy = "") : RuleDct(u, v, w0, w1), m_awy(awy) {}
		RuleRoute(const RuleDct& dct, const Glib::ustring& awy = "") : RuleDct(dct), m_awy(awy) {}
		const Glib::ustring& get_airway(void) const { return m_awy; }

	protected:
		Glib::ustring m_awy;
	};

	class Rule {
	public:
		typedef enum {
			codetype_mandatory = RuleResult::codetype_mandatory,
			codetype_forbidden = RuleResult::codetype_forbidden,
			codetype_closed = RuleResult::codetype_closed,
			codetype_invalid
		} codetype_t;

		Rule(xmlpp::Document *doc);
		bool is_valid(void) const;
		void next_rule(void);
		codetype_t get_codetype(void) const { return m_codetype; }
		static char get_codetype_char(codetype_t t);
		char get_codetype_char(void) const { return get_codetype_char(get_codetype()); }
		const Glib::ustring& get_codeid(void) const { return m_codeid; }
		const Glib::ustring& get_oprgoal(void) const { return m_oprgoal; }
		const Glib::ustring& get_descr(void) const { return m_descr; }
		xmlpp::Element *get_condition(void) const { return m_condel; }
		xmlpp::Element *get_element(void) const { return m_el; }

	protected:
		xmlpp::Element *m_el;
		xmlpp::Element *m_condel;
		Glib::ustring m_codeid;
		Glib::ustring m_oprgoal;
		Glib::ustring m_descr;
		codetype_t m_codetype;

		void parse(void);
	};

	TrafficFlowRestrictions(void);
	bool load(const Glib::ustring& fname);
	Result check_fplan(const FPlanRoute& route, char type_of_flight,
			   const std::string& acft_type, const std::string& equipment,
			   AirportsDb& airportdb, NavaidsDb& navaiddb, WaypointsDb& waypointdb,
			   AirwaysDb& airwaydb, AirspacesDb& airspacedb);

	typedef tracerules_t::const_iterator tracerules_const_iterator_t;
	tracerules_const_iterator_t tracerules_begin(void) const { return m_tracerules.begin(); }
	tracerules_const_iterator_t tracerules_end(void) const { return m_tracerules.end(); }
	bool tracerules_empty(void) const { return m_tracerules.empty(); }
	bool tracerules_exists(const Glib::ustring& ruleid) const;
	bool tracerules_add(const Glib::ustring& ruleid);
	bool tracerules_erase(const Glib::ustring& ruleid);

	Rule get_first_rule(void) { return Rule(m_dom.get_document()); }

protected:
	class CondResult {
	public:
		CondResult(boost::logic::tribool result = boost::logic::indeterminate) : m_result(result) {}
		boost::logic::tribool get_result(void) const { return m_result; }
		void set_result(boost::logic::tribool result = boost::logic::indeterminate) { m_result = result; }
		typedef std::set<unsigned int> set_t;
		set_t& get_vertexset(void) { return m_vertexset; }
		const set_t& get_vertexset(void) const { return m_vertexset; }
		set_t& get_edgeset(void) { return m_edgeset; }
		const set_t& get_edgeset(void) const { return m_edgeset; }
		unsigned int get_first(void) const;
		unsigned int get_last(void) const;
		CondResult& operator&=(const CondResult& x);
		CondResult& operator|=(const CondResult& x);
		CondResult operator~(void);
		CondResult operator&(const CondResult& x) { CondResult r(*this); r &= x; return r; }
		CondResult operator|(const CondResult& x) { CondResult r(*this); r |= x; return r; }

	protected:
		set_t m_vertexset;
		set_t m_edgeset;
		boost::logic::tribool m_result;
	};

	class CompareAirspaces {
	public:
		bool operator()(const AirspacesDb::Airspace& a, const AirspacesDb::Airspace& b);
	};

	xmlpp::DomParser m_dom;
	Fpl m_route;
	AirportsDb *m_airportdb;
	NavaidsDb *m_navaiddb;
	WaypointsDb *m_waypointdb;
	AirwaysDb *m_airwaydb;
	AirspacesDb *m_airspacedb;
	Graph m_graph;
	Rect m_graphbbox;
	typedef std::set<AirspacesDb::Airspace, CompareAirspaces> airspacecache_t;
	airspacecache_t m_airspacecache;
	tracerules_t m_tracerules;
	typedef std::vector<RuleResult> rules_t;
	rules_t m_rules;
	typedef std::vector<Message> messages_t;
	messages_t m_messages;
	std::string m_aircrafttype;
	std::string m_aircrafttypeclass;
	std::string m_equipment;
	char m_typeofflight;
	bool m_shortcircuit;

	bool get_rulewpt(Graph::vertex_descriptor& u, RuleWpt& wpt, xmlpp::Element *el, const Glib::ustring& suffix = "Spn");
	bool get_rulewpt(Graph::vertex_descriptor& u, xmlpp::Element *el, const Glib::ustring& suffix = "Spn");
	static RuleWpt get_rulewpt(xmlpp::Element *el, const Glib::ustring& suffix = "Spn");
	static AltRange get_rulealtrange(xmlpp::Element *el, const Glib::ustring& nodename = "Fcl");
	const AirspacesDb::Airspace& get_ruleairspace(xmlpp::Element *el, const Glib::ustring& nodename = "AseUid");
	bool get_ruledct(RuleDct& dct, xmlpp::Element *el);
	bool get_ruleawy(RuleRoute& rte, xmlpp::Element *el);
	bool check_fplan_depdest(void);
	bool check_fplan_rules(void);
	void load_airway_graph(void);
	CondResult compute_condition_1(Rule& rule, xmlpp::Element *cond);
	CondResult compute_condition(Rule& rule, xmlpp::Element *cond);
	CondResult compute_tfe_1(Rule& rule, xmlpp::Element *tfe, RuleResult::Alternative::Sequence& seq);
	CondResult compute_tfe(Rule& rule, xmlpp::Element *tfe, RuleResult::Alternative::Sequence& seq);
};

#endif /* TFR_H */
